/* 
   Copyright (C) 2000 Robin Hogan

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <cstdio>
#include <cmath>
#include "rplot.h"
#define AXES_INTERP(x1,x2,x,w1,w2) ((w1) + (((w2)-(w1))*((x)-(x1)))/((x2)-(x1)))
#define AXES_PRETNI(x1,x2,w,w1,w2) ((x1) + (((x2)-(x1))*((w)-(w1)))/((w2)-(w1)))

// *** RAxes class ***

void
RAxes::init(RFigure *parent, RRectangle pos)
{
  //	_format=CHILD;
  _parent=parent;
  /*
    _xoffset=INTERP1(parent->_axis.x1,parent->_axis.x2,MIN(pos.x1,pos.x2),
    _parent._xoffset,_parent._width);
    _yoffset=INTERP1(parent->_axis.y1,parent->_axis.y2,MIN(pos.y1,pos.y2),
    _parent._yxoffset,_parent._height);
    _width=INTERP1(parent->_axis.x1,parent->_axis.x2,MAX(pos.x1,pos.x2),
    _parent._xoffset,_parent._width)-_width;
    _height=INTERP1(parent->_axis.y1,parent->_axis.y2,MAX(pos.y1,pos.y2),
    _parent._yxoffset,_parent._height)-_height;
  */
  _position=pos;
  _extent=pos;
  _clipbox=make_rectangle(0,0,1,1);
  axis(_clipbox);
  _clipping=0;
  _pen_color=_color=make_color(0,0,0);
  _line_type=LINE_SOLID;
}

void
RAxes::pixel(double x, double y, const RColor &col)
{
  if (_clipping && !_clipbox.in_strict(x,y)) {
    return;
  }
  x=AXES_INTERP(_axis.x1,_axis.x2,x,_position.x1,_position.x2);
  y=AXES_INTERP(_axis.y1,_axis.y2,y,_position.x1,_position.x2);
  _parent->pixel(x, y, col);
}

void
RAxes::plot(double x1, double y1, double x2, double y2, const RColor &col,
	    LINE line_type)
{
  x1=AXES_INTERP(_axis.x1,_axis.x2,x1,_position.x1,_position.x2);
  y1=AXES_INTERP(_axis.y1,_axis.y2,y1,_position.y1,_position.y2);
  x2=AXES_INTERP(_axis.x1,_axis.x2,x2,_position.x1,_position.x2);
  y2=AXES_INTERP(_axis.y1,_axis.y2,y2,_position.y1,_position.y2);

  bool fig_clipping = _parent->clipping();
  if (_clipping) {
    _parent->clipbox(_position);
  }
  _parent->plot(x1, y1, x2, y2, col, line_type);
  if (_clipping) {
    _parent->restore_clipbox();
    _parent->clipbox(fig_clipping);
  }
}

void
RAxes::barb(double x, double y, double dx, double dy, const RColor &col,
	    LINE line_type, double *xreturn, double *yreturn)
{
  double xret, yret;

  x=AXES_INTERP(_axis.x1,_axis.x2,x,_position.x1,_position.x2);
  y=AXES_INTERP(_axis.y1,_axis.y2,y,_position.y1,_position.y2);
  _parent->barb(x, y, dx, dy, col, line_type, &xret, &yret);

  if (xreturn != NULL) {
    (*xreturn) = AXES_PRETNI(_axis.x1,_axis.x2,xret,
			     _position.x1,_position.x2);
  }
  if (yreturn != NULL) {
    (*yreturn) = AXES_PRETNI(_axis.y1,_axis.y2,yret,
			     _position.y1,_position.y2);
  }
}

void
RAxes::clear(const RColor &col)
{
  _parent->rectangle(_position.x1, _position.y1,
		     _position.x2, _position.y2, col);
}

void
RAxes::rectangle(double x1, double y1, double x2, double y2,
		 const RColor &col)
{
  double minx,maxx,miny,maxy;

  if (_clipping) {
    if (x1>x2) { minx=x2;maxx=x1; } else { minx=x1;maxx=x2; };
    if (y1>y2) { miny=y2;maxy=y1; } else { miny=y1;maxy=y2; };
    x1=MAX(minx,_clipbox.x1);
    x2=MIN(maxx,_clipbox.x2);
    y1=MAX(miny,_clipbox.y1);
    y2=MIN(maxy,_clipbox.y2);
    if (x1>x2 || y1>y2) {
      return;
    }
  }
  minx=AXES_INTERP(_axis.x1,_axis.x2,x1,_position.x1,_position.x2);
  maxx=AXES_INTERP(_axis.x1,_axis.x2,x2,_position.x1,_position.x2);
  miny=AXES_INTERP(_axis.y1,_axis.y2,y1,_position.y1,_position.y2);
  maxy=AXES_INTERP(_axis.y1,_axis.y2,y2,_position.y1,_position.y2);

  _parent->rectangle(minx, miny, maxx, maxy, col);
}

RRectangle
RAxes::text(double x, double y, const char *str, ANCHOR anchor, int rotation,
	    RFont *fnt, const RColor &col)
{
  RRectangle rect;

  x=AXES_INTERP(_axis.x1,_axis.x2,x,_position.x1,_position.x2);
  y=AXES_INTERP(_axis.y1,_axis.y2,y,_position.y1,_position.y2);
  rect=_parent->text(x,y,str,anchor,rotation,fnt,col);
  return make_rectangle(
			AXES_INTERP(_position.x1,_position.x2,rect.x1,_axis.x1,_axis.x2),
			AXES_INTERP(_position.y1,_position.y2,rect.y1,_axis.y1,_axis.y2),
			AXES_INTERP(_position.x1,_position.x2,rect.x2,_axis.x1,_axis.x2),
			AXES_INTERP(_position.y1,_position.y2,rect.y2,_axis.y1,_axis.y2)
			);
}

void
RAxes::xlabel(char *str, bool otherside, RFont *fnt)
{
  if (!otherside) {
    _parent->text(0.5 * (_position.x1+_position.x2), _extent.y1, str,
		  ANCHOR_TOP, 0, fnt, _pen_color);
  }
  else  {
    _parent->text(0.5 * (_position.x1+_position.x2), _extent.y2, str,
		  ANCHOR_BOTTOM, 0, fnt, _pen_color);
  }
}

void
RAxes::ylabel(char *str, bool otherside, RFont *fnt)
{
  if (!otherside) {
    _parent->text(_extent.x1, 0.5 * (_position.y1+_position.y2), str,
		  ANCHOR_BOTTOM, 90, fnt, _pen_color);
  }
  else  {
    _parent->text(_extent.x2, 0.5 * (_position.y1+_position.y2), str,
		  ANCHOR_TOP, 90, fnt, _pen_color);
  }
}

/*
void RAxes::xticks(double length, bool opposite, bool grid) {};
void RAxes::yticks(double length, bool opposite, bool grid);
*/

void
RAxes::xticks(double start, double interval, char *format, double length, 
	      bool opposite, bool grid, bool singleticks)
{
  double x,min=MIN(_axis.x1,_axis.x2),max=MAX(_axis.x1,_axis.x2);
  char buf[32];
  if (format == NULL) {
    format = "%g";
  }
  if (interval <= 0) {
    interval=max-min;
  }
  for (x=start;x<=max;x+=interval) {
    if (x>=min) {
      sprintf((char *) buf,format,x);
      xtick(x,buf,length,opposite,grid,singleticks);
    }
  }
}

void
RAxes::xticks_time(double start, double interval, double length, 
		   bool opposite, bool grid, bool singleticks)
{
  double x,minx=MIN(_axis.x1,_axis.x2),maxx=MAX(_axis.x1,_axis.x2);
  char buf[32];
  char *format="%02d:%02d:%02.2d";
  int hour=0, min=0, sec=0;
  if (interval <= 0) {
    interval=maxx-minx;
  }
  if ((start != (int) start) || (interval != (int) interval)) {
    if ((start*60.0 == (int) (start*60.0)) && 
	(interval*60.0 == (int) (interval*60.0))) {
      format="%02d:%02d";
    }
  }
  else {
    format="%02d:00";
  }
  for (x=start;x<=maxx;x+=interval) {
    if (x>=minx) {
      hour=(int) x;
      min=(int) (60.0*(x-hour));
      sec=(int) (60.0 * ((60.0*(x-hour))-min));
      if (hour >= 0) {
	hour %= 24;
      }
      else {
	hour = 24-( (-hour) % 24);
      }
      sprintf((char *) buf,format,hour,min,sec);
      xtick(x,buf,length,opposite,grid,singleticks);
    }
  }
}

void
RAxes::yticks(double start, double interval, char *format, double length, 
	      bool opposite, bool grid, bool singleticks)
{
  double y,min=MIN(_axis.y1,_axis.y2),max=MAX(_axis.y1,_axis.y2);
  char buf[32];

  if (format == NULL) {
    format = "%g";
  }
  if (interval <= 0) {
    interval=max-min;
  }
  for (y=start;y<=max;y+=interval) {
    if (y>=min) {
      if (fabs(y) < 0.01*interval) {
	/* y probably should be exactly zero here */
	y = 0.0;
      }
      sprintf((char *) buf,format,y);
      ytick(y,buf,length,opposite,grid,singleticks);
    }
  }
}

void
RAxes::xtick(double x, char *str, double length, bool opposite, 
	     bool grid, bool singleticks)
{
  RRectangle rect;
  double y, dy;
  double yret1, yret2;

  x = AXES_INTERP(_axis.x1,_axis.x2,x,_position.x1,_position.x2);

  if (grid) {
    _parent->plot(x,_position.y1,x,_position.y2,_color,LINE_DOTTED);
  }

  dy = length; // * ABS(_position.y2-_position.y1);
	
  if (opposite) {
    if (length != 0.0) {
      // _parent->plot(x,_position.y2,x,_position.y2-dy,_color);
      _parent->barb(x,_position.y2,0,-dy,_color,LINE_SOLID,NULL,&yret2);
      if (!singleticks) {
	// _parent->plot(x,_position.y1,x,_position.y1+dy,_color);
	_parent->barb(x,_position.y1,0,dy,_color,LINE_SOLID,NULL,&yret1);
      }
    }
    // y = (length<0 ? _position.y2-dy : _position.y2);
    y = (length<0 ? yret2 : _position.y2);
    if (str != NULL) {
      if (str[0] != '\0') {
	rect=_parent->text(x,y,str,ANCHOR_BOTTOM,0,NULL,_pen_color);
	_extent.y2=MAX(_extent.y2,rect.y2);
      }
    }
  }
  else {
    if (length != 0.0) {
      // _parent->plot(x,_position.y1,x,_position.y1+dy,_color);
      _parent->barb(x,_position.y1,0,dy,_color,LINE_SOLID,NULL,&yret1);
      if (!singleticks) {
	// _parent->plot(x,_position.y2,x,_position.y2-dy,_color);
	_parent->barb(x,_position.y2,0,-dy,_color,LINE_SOLID,NULL,&yret2);
      }
    }
    // y = (length<0 ? _position.y1+dy : _position.y1);
    y = (length<0 ? yret1 : _position.y1);
    if (str != NULL) {
      if (str[0] != '\0') {
	rect=_parent->text(x,y,str,ANCHOR_TOP,0,NULL,_pen_color);
	_extent.y1=MIN(_extent.y1,rect.y1);
      }
    }
  }
}

void
RAxes::ytick(double y, char *str, double length, bool opposite, 
	     bool grid, bool singleticks)
{
  RRectangle rect;
  double x, dx;
  double xret1, xret2;

  y = AXES_INTERP(_axis.y1,_axis.y2,y,_position.y1,_position.y2);

  if (grid) {
    _parent->plot(_position.x1,y,_position.x2,y,_color,LINE_DOTTED);
  }

  dx = length; // * ABS(_position.x2-_position.x1);
	
  if (opposite) {
    if (length != 0.0) {
      // _parent->plot(_position.x2,y,_position.x2-dx,y,_color);
      _parent->barb(_position.x2,y,-dx,0,_color,LINE_SOLID,&xret2,NULL);
      if (!singleticks) {
	// _parent->plot(_position.x1,y,_position.x1+dx,y,_color);
	_parent->barb(_position.x1,y,dx,0,_color,LINE_SOLID,&xret1,NULL);
      }
    }
    // x = (length<0 ? _position.x2-dx : _position.x2);
    x = (length<0 ? xret2 : _position.x2);
    if (str != NULL) {
      if (str[0] != '\0') {
	rect=_parent->text(x,y,str,ANCHOR_LEFT,0,NULL,_pen_color);
	_extent.x2=MAX(_extent.x2,rect.x2);
      }
    }
  }
  else {
    if (length != 0.0) {
      // _parent->plot(_position.x1,y,_position.x1+dx,y,_color);
      _parent->barb(_position.x1,y,dx,0,_color,LINE_SOLID,&xret1,NULL);
      if (!singleticks) {
	// _parent->plot(_position.x2,y,_position.x2-dx,y,_color);
	_parent->barb(_position.x2,y,-dx,0,_color,LINE_SOLID,&xret2,NULL);
      }
    }
    // x = (length<0 ? _position.x1+dx : _position.x1);
    x = (length<0 ? xret1 : _position.x1);
    if (str != NULL) {
      if (str[0] != '\0') {
	rect=_parent->text(x,y,str,ANCHOR_RIGHT,0,NULL,_pen_color);
	_extent.x1=MIN(_extent.x1,rect.x1);
      }
    }
  }
}



void
RAxes::polygon(int n, float *x, float *y, const RColor &col)
{
  float xx[RPLOT_MAX_POLYGON_VERTICES], yy[RPLOT_MAX_POLYGON_VERTICES];

  bool fig_clipping = _parent->clipping();

  if (n < 3) {
    throw("Can't draw polygon with fewer than 3 vertices");
  }
  else if (n > RPLOT_MAX_POLYGON_VERTICES) {
    throw("Too many vertices requested for polygon");
  }

  //  xx = new float[2*n]; // Too slow?
  //  yy = xx + n;

  for (int i = 0; i < n; ++i) {
    xx[i] = AXES_INTERP(_axis.x1,_axis.x2,x[i],_position.x1,_position.x2);
    yy[i] = AXES_INTERP(_axis.y1,_axis.y2,y[i],_position.y1,_position.y2);
  }
  if (_clipping) {
    _parent->clipbox(_position);
  }
  _parent->polygon(n, xx, yy, col);
  if (_clipping) {
    _parent->restore_clipbox();
    _parent->clipbox(fig_clipping);
  }
  // delete[] xx;
  return;
}
