/* 
   Copyright (C) 2000 Robin Hogan

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "rplot.h"

void
RAxes::hray(float *x, float y1, float y2, float *val, int n, RPalette *pal,
		 float minval, float maxval, float nanval)
{
  int i, j, size;
  float dval=maxval-minval;

  if (! (pal->contiguous())) {
    throw("RPalette is not contiguous");
  }
  if ((size = pal->size()) == 0) {
    throw("RPalette is empty");
  }
  if (dval == 0) {
    throw("No color range");
  }

  for (i=0;i<n-1;i++) {
    if (val[i] != nanval) {
      j = (int) (((float) size) * (val[i]-minval) / dval);
      if (j < 0) j = 0;
      if (j > size-1) j = size-1;
      rectangle(x[i],y1,x[i+1],y2,pal->get(j));
    }
  }
}

void
RAxes::vray(float x1, float x2, float *y, float *val, int n, RPalette *pal,
		 float minval, float maxval, float nanval)
{
  int i, j, size;
  float dval=maxval-minval;

  if (! (pal->contiguous())) {
    throw("RPalette is not contiguous");
  }
  if ((size = pal->size()) == 0) {
    throw("RPalette is empty");
  }
  if (dval == 0) {
    throw("No color range");
  }

  for (i=0;i<n-1;i++) {
    if (val[i] != nanval) {
      j = (int) (((float) size) * (val[i]-minval) / dval);
      if (j < 0) j = 0;
      if (j > size-1) j = size-1;
      rectangle(x1,y[i],x2,y[i+1],pal->get(j));
    }
  }
}

void
RAxes::colorbar(RPalette *pal,ORIENTATION orient, int size)
{
  int i, true_size;
  double dd;

  if (! (pal->contiguous())) {
    throw("RPalette is not contiguous");
  }
  if ((true_size = pal->size()) == 0) {
    throw("RPalette is empty");
  }

  if (size < 1) {
    size = true_size;
  }

  if (orient == HORIZONTAL) {
    dd=_axis.x2-_axis.x1;
    for (i=0;i<size;i++) {
      rectangle(_axis.x1 + (i*dd)/size,_axis.y1,
		_axis.x1 + ((i+1)*dd)/size,_axis.y2, pal->get(i) );
    }
  }
  else {
    dd=_axis.y2-_axis.y1;
    for (i=0;i<size;i++) {
      rectangle(_axis.x1, _axis.y1 + (i*dd)/size,
		_axis.x2, _axis.y1 + ((i+1)*dd)/size, pal->get(i) );
    }
  }
}



