#include <string.h>
#include <stdio.h>
#include <netcdf.h>
#include "zdr_phidp.h"
#include "oe_netcdf.h"

int
check_vector_length(OEvector *v, int n)
{
  if (v->n != n) {
    OE_THROW(OE_DIMENSION_ERROR, "Scan vector is the wrong length");
  }
  else {
    return OE_SUCCESS;
  }
}

int
check_matrix_size(OEmatrix *M, int m, int n)
{
  if (M->m != m || M->n != n) {
    OE_THROW(OE_DIMENSION_ERROR, "Scan matrix is the wrong size");
  }
  else {
    return OE_SUCCESS;
  }
}

int
load_scan(char *nc_filename, Scan *scan)
{
  int ncid, timedimid, rangedimid;
  int nrays, ngates;
  int varid;
  NC_CHECK(nc_open(nc_filename, NC_NOWRITE, &ncid));
  NC_CHECK(nc_inq_dimid(ncid, "time", &timedimid));
  NC_CHECK(nc_inq_dimlen(ncid, timedimid, &nrays));
  NC_CHECK(nc_inq_dimid(ncid, "range", &rangedimid));
  NC_CHECK(nc_inq_dimlen(ncid, rangedimid, &ngates));

  OE_CHECK(oe_read_nc_vector(ncid, "range", &scan->range));
  OE_CHECK(check_vector_length(&scan->range, ngates));

  OE_CHECK(oe_read_nc_vector(ncid, "azim", &scan->azim));
  OE_CHECK(check_vector_length(&scan->azim, nrays));

  OE_CHECK(oe_read_nc_vector(ncid, "elev", &scan->elev));
  OE_CHECK(check_vector_length(&scan->elev, nrays));

  OE_CHECK(oe_read_nc_vector(ncid, "time", &scan->time));
  OE_CHECK(check_vector_length(&scan->time, nrays));

  OE_CHECK(oe_read_nc_matrix(ncid, "Zh", &scan->Zh));
  OE_CHECK(check_matrix_size(&scan->Zh, nrays, ngates));
  if (nc_inq_varid(ncid, "Zh_error", &varid) == NC_NOERR) {
    OE_CHECK(oe_read_nc_matrix(ncid, "Zh_error", &scan->Zh_error));
    OE_CHECK(check_matrix_size(&scan->Zh_error, nrays, ngates));
    scan->is_Zh_error = 1;
  }
  else {
    scan->is_Zh_error = 0;
  }

  OE_CHECK(oe_read_nc_matrix(ncid, "temperature", &scan->temperature));
  OE_CHECK(check_matrix_size(&scan->temperature, nrays, ngates));

  if (nc_inq_varid(ncid, "Zdr", &varid) == NC_NOERR) {
    OE_CHECK(oe_read_nc_matrix(ncid, "Zdr", &scan->Zdr));
    OE_CHECK(check_matrix_size(&scan->Zdr, nrays, ngates));
    OE_CHECK(oe_read_nc_matrix(ncid, "Zdr_error", &scan->Zdr_error));
    OE_CHECK(check_matrix_size(&scan->Zdr_error, nrays, ngates));
    scan->is_Zdr = 1;
  }
  else {
    scan->is_Zdr = 0;
  }

  if (nc_inq_varid(ncid, "phi_dp", &varid) == NC_NOERR) {
    OE_CHECK(oe_read_nc_matrix(ncid, "phi_dp", &scan->phi_dp));
    OE_CHECK(check_matrix_size(&scan->phi_dp, nrays, ngates));
    OE_CHECK(oe_read_nc_matrix(ncid, "phi_dp_error", &scan->phi_dp_error));
    OE_CHECK(check_matrix_size(&scan->phi_dp_error, nrays, ngates));
    scan->is_phi_dp = 1;
  }
  else {
    scan->is_phi_dp = 0;
  }

  if (!scan->is_Zdr || !scan->is_phi_dp) {
    OE_THROW(OE_FAILURE, "Scan file contains neither Zdr nor phi_dp");
  }

  OE_CHECK(oe_read_nc_scalar(ncid, "frequency", &scan->frequency));
  NC_CHECK(nc_close(ncid));

  return OE_SUCCESS;
}
