#include <stdio.h>
#include <math.h>
#include "zdr_phidp.h"

#define A_PRIORI 0
#define KALMAN_FORWARD 1
#define KALMAN_BACKWARD 2

//#define INVERT oe_invert
//#define SOLVE oe_solve
#define INVERT oe_invert_symmetric
#define SOLVE oe_solve_symmetric
#define PI_OVER_180 0.0174532925199433 

static
int
add_spatial_decorrelation(OEmatrix *target, OEreal azim_decorr_dist,
			  OEreal var_ln_a_apriori,
			  OEvector *range, OEreal azim1, OEreal azim2)
{
  OEreal drange, range0, dazim = fabs(azim1 - azim2);
  OEreal azim_decorr_factor = 1.0/azim_decorr_dist;
  int i, n = target->n;
  if (n != target->m) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Spatial decorrelation covariance matrix must be square");
  }
  if (dazim > 360.0) {
    dazim = 360.0 - dazim;
  }
  dazim *= PI_OVER_180;

  drange = (range->data[range->n-1]-range->data[0]) / n;
  range0 = range->data[0] + 0.5*drange;

  /*  fprintf(stderr, "%g %g %g %g\n", drange, range0, dazim, azim_decorr_factor);*/

  /*  azim_decorr_factor /= 1.0;*/

  for (i = 0; i < n; i++) {
    /* target->data[i][i] += azim_decorr_factor * dazim * (range0+drange*i);*/
    //    target->data[i][i] += (target->data[i][i]+var_ln_a_apriori)
    //      * azim_decorr_factor * dazim * (range0+drange*i);
    /* Really we should multiply by obs error in x: KT_invSy_K */
    target->data[i][i] += var_ln_a_apriori
      * azim_decorr_factor * dazim * (range0+drange*i);
  }

  return OE_SUCCESS;
}

int
zdr_phidp(Settings *settings, ScatteringLUT *lut, Scan *scan,
	  Retrieval *retrieval)
{
  OEvector range = OE_NEW_VECTOR;
  OEvector Zh_lin = OE_NEW_VECTOR;
  OEvector temperature = OE_NEW_VECTOR;

  OEvector y = OE_NEW_VECTOR;
  OEvector inv_sy = OE_NEW_VECTOR;

  OEvector y_fwd = OE_NEW_VECTOR;
  OEvector Zdr_fwd = OE_NEW_VECTOR;
  OEvector phi_dp_fwd = OE_NEW_VECTOR;

  /* The state vector */
  OEvector x = OE_NEW_VECTOR;
  /* The first part of x: basis function ln(a) */
  OEvector ln_a_basis = OE_NEW_VECTOR;
  /* The second part of x: hail fractions in selected gates */
  OEvector hail_fraction = OE_NEW_VECTOR;
  /* The values of ln(a) at rainy gates */
  OEvector ln_a = OE_NEW_VECTOR;
  /* The values of ln(a) at all gates */
  OEvector ln_a_full = OE_NEW_VECTOR;

  OEvector xa = OE_NEW_VECTOR;
  OEvector inv_sa = OE_NEW_VECTOR;
  OEmatrix inv_Sa = OE_NEW_MATRIX;
  OEmatrix Sa = OE_NEW_MATRIX;
  OEmatrix Slna = OE_NEW_MATRIX;

  OEmatrix W_full = OE_NEW_MATRIX;
  OEmatrix W = OE_NEW_MATRIX;

  OEvector Ah = OE_NEW_VECTOR;
  OEvector Av = OE_NEW_VECTOR;

  OEmatrix Jacobian = OE_NEW_MATRIX;
  OEmatrix Jacobian_highres = OE_NEW_MATRIX;
  OEmatrix Jacobian_rainonly = OE_NEW_MATRIX;

  OEmatrix dZdr_dlna = OE_NEW_MATRIX;
  OEmatrix dphi_dp_dlna = OE_NEW_MATRIX;

  OEmatrix Hessian = OE_NEW_MATRIX;
  OEmatrix invHessian = OE_NEW_MATRIX;
  OEmatrix Sa_true = OE_NEW_MATRIX;
  OEvector rhs = OE_NEW_VECTOR;
  OEvector sp = OE_NEW_VECTOR;

  OEmatrix dAh_dlna = OE_NEW_MATRIX;
  OEmatrix dAv_dlna = OE_NEW_MATRIX;
  OEmatrix dAh_dhf = OE_NEW_MATRIX;
  OEmatrix dAv_dhf = OE_NEW_MATRIX;
  OEmatrix dZdr_dhf = OE_NEW_MATRIX;
  OEmatrix dphi_dp_dhf = OE_NEW_MATRIX;

  OEvector dZdr_das = OE_NEW_VECTOR;
  OEvector dphi_dp_das = OE_NEW_VECTOR;
  OEvector dAh_das = OE_NEW_VECTOR;
  OEvector dAv_das = OE_NEW_VECTOR;

  OEmatrix lna_save = OE_NEW_MATRIX;
  OEmatrix Slna_save = OE_NEW_MATRIX;
  OEmatrix invSkalman = OE_NEW_MATRIX;
  OEvector xkalman = OE_NEW_VECTOR;

  OEvector lnR_error = OE_NEW_VECTOR;

  OEintvector gate_list = OE_NEW_VECTOR;
  OEintvector hail_gate_list = OE_NEW_VECTOR;

  OEreal ln_a_default = log(settings->a_default);

  OEreal last_azim = 0;

  OEreal sec_fm = 0.0, sec_prep = 0.0, sec_solve = 0.0;
  OEreal sec_spline1 = 0.0, sec_spline2 = 0.0;
  OEreal sec_reset = 0.0, sec_save = 0.0;

  int iray = 0;
  int ngates = scan->range.n;
  int nmaxhailgates = ngates;
  int nrays = scan->Zh.m;
  int ix;
  int nb = settings->n_basis;
  int na = (settings->var_atten_scaling > 0.0);
  int ipass = 0;
  int npass = (settings->azim_decorr_dist > 0.0) + 1;

  /* Set this variable to a positive number to save the matrices from
     a particular ray in ascii files */
  int iray_save = -1;

  OEreal drange = scan->range.data[1]-scan->range.data[0];

  OEreal chi_squared_Zdr = 0.0;
  OEreal chi_squared_phi_dp = 0.0;
  int n_total = 0;
  OEreal atten_scaling = settings->atten_scaling;

  if (scan->is_Zdr && scan->is_phi_dp) {
    fprintf(stderr, "Zdr/Phidp retrieval");
  }
  else if (scan->is_Zdr) {
    fprintf(stderr, "Zdr-only retrieval");
  }
  else {
    fprintf(stderr, "Phidp-only retrieval");
  }
  if (settings->do_hail_retrieval) {
    fprintf(stderr, " with hail: %d rays, %d gates\n",
	    nrays, ngates);
  }
  else {
    fprintf(stderr, " without hail: %d rays, %d gates\n",
	    nrays, ngates);
    nmaxhailgates = 0;
  }
  OE_CHECK(oe_cubic_b_spline(&W_full, ngates, nb));

  /* Reset the output matrices */
  OE_CHECK(oe_reset_matrix(&retrieval->Zdr_fwd, nrays, ngates));
  OE_CHECK(oe_reset_matrix(&retrieval->phi_dp_fwd, nrays, ngates));
  OE_CHECK(oe_reset_matrix(&retrieval->R, nrays, ngates));
  OE_CHECK(oe_reset_matrix(&retrieval->lnR_error, nrays, ngates));
  OE_CHECK(oe_reset_matrix(&retrieval->Ah, nrays, ngates));
  OE_CHECK(oe_reset_matrix(&retrieval->Av, nrays, ngates));
  OE_CHECK(oe_reset_matrix(&retrieval->a, nrays, ngates));
  OE_CHECK(oe_reset_matrix(&retrieval->lna_error, nrays, ngates));
  if (settings->do_hail_retrieval) {
    OE_CHECK(oe_reset_matrix(&retrieval->hail_fraction, nrays, ngates));
    OE_CHECK(oe_zero_matrix(&retrieval->hail_fraction));
    OE_CHECK(oe_reset_matrix(&retrieval->hail_fraction_error, nrays, ngates));
    OE_CHECK(oe_zero_matrix(&retrieval->hail_fraction_error));
  }
  if (na) {
    OE_CHECK(oe_reset_vector(&retrieval->atten_scaling, nrays));
    OE_CHECK(oe_reset_vector(&retrieval->atten_scaling_error, nrays));
  }

  /* Set the a priori and its error variance */
  OE_CHECK(oe_reset_vector(&xa, nb+na+nmaxhailgates));
  if (settings->explicit_range_smoothing) {
    OE_CHECK(oe_reset_vector(&inv_sa, nb+na+nmaxhailgates));
  }
  else {
    OE_CHECK(oe_reset_matrix(&Sa, nb+na+nmaxhailgates,
			     nb+na+nmaxhailgates));
    OE_CHECK(oe_reset_matrix(&inv_Sa, nb+na+nmaxhailgates,
			     nb+na+nmaxhailgates));
    OE_CHECK(oe_zero_matrix(&Sa));
    OE_CHECK(oe_link_matrix(&Slna, nb, nb, Sa.data));
  }

  /* Elements 0 to nb-1 are the ln(a) values */
  for (ix = 0; ix < nb; ix++) {
    xa.data[ix] = ln_a_default;
  }
  if (settings->explicit_range_smoothing) {
    for (ix = 0; ix < nb; ix++) {
      inv_sa.data[ix] = 1.0 / settings->var_ln_a_apriori;
    }
  }
  else {
    OEreal radial_decorr_factor = settings->radial_decorr_dist*nb
      /(drange*ngates); 
    OE_CHECK(oe_exponential_errcov(nb, settings->var_ln_a_apriori,
				   radial_decorr_factor, &Slna));
  }
  /* Elements nb to nb+na are for attenuation scaling retrieval */
  if (na) {
    xa.data[ix] = settings->atten_scaling;
    if (settings->explicit_range_smoothing) {
      inv_sa.data[ix] = 1.0 / settings->var_atten_scaling;
    }
    else {
      Sa.data[ix][ix] = settings->var_atten_scaling;
    }
    ix++;
  }
  /* Elements nb+na to nb+na+nmaxhailgates-1 are for hail - weak zero
     a priori here */
  if (settings->explicit_range_smoothing) {
    for ( ; ix < nb+na+nmaxhailgates; ix++) {
      xa.data[ix] = 0.0;
      inv_sa.data[ix] = 1.0/(2.0*2.0);
    }
  }
  else {
    for ( ; ix < nb+na+nmaxhailgates; ix++) {
      xa.data[ix] = 0.0;
      Sa.data[ix][ix] = 2.0*2.0;
    }
    OE_CHECK(INVERT(&Sa, &inv_Sa));
    oe_free_matrix(&Sa);
  }

  /* If using a Kalman filter, initialise the matrices to save the
     solution and its error covariance at each ray */
  if (npass > 1) {
    OE_CHECK(oe_reset_matrix(&lna_save, nrays, nb+na));
    OE_CHECK(oe_reset_matrix(&Slna_save, nrays*(nb+na), nb+na));
  }

  OE_CHECK(oe_reset_int_vector(&gate_list, ngates));

  /* Loop through each pass and each ray */
  if (npass > 1) {
    fprintf(stderr, "Forward pass ");
  }

  while (ipass < npass) {
    OEreal last_Ah = 0.0, last_Av = 0.0, last_phi_dp = 0.0;
    int igate = 0;
    int iraingate, nraingates = 0;
    int nx, ny;
    int iter, iter_max = 5;
    int ihailgate = 0, nhailgates = 0;

    oe_time_split();

    /* Find the gates containing rain, i.e. finite values in Zh, Zdr
       and phi_dp */
    while (igate < ngates) {
      if (scan->Zh.data[iray][igate] == scan->Zh.missing_value
	  || (scan->is_Zdr 
	      && scan->Zdr.data[iray][igate]
	      == scan->Zdr.missing_value)
	  || (scan->is_phi_dp
	      && scan->phi_dp.data[iray][igate]
	      == scan->phi_dp.missing_value)) {
	igate++;
      }
      else {
	gate_list.data[nraingates] = igate;
	nraingates++;
	igate++;
      }
    }
    gate_list.n = nraingates;

    if (nraingates <= 0) {
      OE_THROW(OE_FAILURE, "Ray found containing no rain: a bit of a limitation but I can't cope with such rays!");
      continue;
    }

    /* Hail stuff */
    if (settings->do_hail_retrieval) {
      if (!retrieval->hail_location.data) {
	OE_THROW(OE_FAILURE,
	 "Hail location must be set if hail retrieval is to be performed");
      }
      /* Number of hail gates for this ray is stored in the first
	 column of hail_location */
      nhailgates = retrieval->hail_location.data[iray][0];
    }
    if (nhailgates) {
      fprintf(stderr, "H");
    }
    else {
      fprintf(stderr, ".");
    }


    /* Our state vector contains the nb entries for the ln(a) basis
       function values, and nhailgates entries containing the fraction
       of the reflectivity that is due to hail at each gate that
       contains hail */
    nx = nb + na + nhailgates;
    /* The a priori must be sized to include null elements for the
       hail properties */
    xa.n = nx;
    if (settings->explicit_range_smoothing) {
      inv_sa.n = nx;
    }
    else {
      inv_Sa.n = inv_Sa.m = nx;
    }

    /* Reset the vectors to the appropriate size */
    OE_CHECK(oe_reset_vector(&range, nraingates));
    OE_CHECK(oe_reset_vector(&Zh_lin, nraingates));
    OE_CHECK(oe_reset_vector(&temperature, nraingates));
    OE_CHECK(oe_reset_vector(&ln_a, nraingates));
    ny = nraingates*(scan->is_Zdr+scan->is_phi_dp);
    OE_CHECK(oe_reset_vector(&y, ny));
    OE_CHECK(oe_reset_vector(&y_fwd, ny));
    OE_CHECK(oe_reset_vector(&inv_sy, ny));
    if (scan->is_Zdr) {
      OE_CHECK(oe_link_vector(&Zdr_fwd, nraingates, y_fwd.data));
    }
    if (scan->is_phi_dp) {
      OE_CHECK(oe_link_vector(&phi_dp_fwd, nraingates,
			      y_fwd.data + scan->is_Zdr * nraingates));
    }

    OE_CHECK(oe_reset_vector(&Ah, nraingates));
    OE_CHECK(oe_reset_vector(&Av, nraingates));
    OE_CHECK(oe_reset_vector(&rhs, nx));

    /* hail_gate_list contains a list of the indeces to the list of
       rainy gates that also contain hail */
    if (nhailgates) {
      OE_CHECK(oe_reset_int_vector(&hail_gate_list, nhailgates));
    }

    /* Fill the vectors */
    for (iraingate = 0; iraingate < nraingates; iraingate++) {
      igate = gate_list.data[iraingate];
      range.data[iraingate] = scan->range.data[igate];
      Zh_lin.data[iraingate] = pow(10.0, 0.1*scan->Zh.data[iray][igate]);
      temperature.data[iraingate] = scan->temperature.data[iray][igate];
      if (scan->is_Zdr) {
	y.data[iraingate] = scan->Zdr.data[iray][igate];
	inv_sy.data[iraingate] 
	  = 1.0 / (scan->Zdr_error.data[iray][igate] 
		   * scan->Zdr_error.data[iray][igate]);
      }
      if (scan->is_phi_dp) {
	y.data[iraingate + nraingates*scan->is_Zdr]
	  = scan->phi_dp.data[iray][igate];
	inv_sy.data[iraingate + nraingates*scan->is_Zdr] 
	  = 1.0 / (scan->phi_dp_error.data[iray][igate] 
		   * scan->phi_dp_error.data[iray][igate]);
      }
      if (ihailgate < nhailgates
	  && igate == retrieval->hail_location.data[iray][ihailgate+1]) {
	hail_gate_list.data[ihailgate++] = iraingate;
      }
    }
    if (nhailgates) {
      if (ihailgate != nhailgates) {
	OE_THROW(OE_FAILURE, "Hail gates must also be rainy gates");
      }
      /* Note that forward model requires hail info at same gates as
	 rain */
      OE_CHECK(oe_reset_vector(&hail_fraction, nraingates));
      OE_CHECK(oe_zero_vector(&hail_fraction));
    }

    /* Reset the matrices to the appropriate size */
    OE_CHECK(oe_reset_matrix(&Jacobian, ny, nx));
    OE_CHECK(oe_reset_matrix(&Jacobian_highres, ny, nraingates));
    OE_CHECK(oe_reset_matrix(&dAh_dlna, nraingates, nraingates));
    OE_CHECK(oe_reset_matrix(&dAv_dlna, nraingates, nraingates));
    if (na) {
      OE_CHECK(oe_reset_vector(&dAh_das, nraingates));
      OE_CHECK(oe_reset_vector(&dAv_das, nraingates));
      if (scan->is_Zdr) {
	OE_CHECK(oe_reset_vector(&dZdr_das, nraingates));
      }
      if (scan->is_phi_dp) {
	OE_CHECK(oe_reset_vector(&dphi_dp_das, nraingates));
      }
    }

    /* Set matrices to zero */    
    OE_CHECK(oe_zero_matrix(&Jacobian_highres));
    OE_CHECK(oe_zero_matrix(&Jacobian));
    OE_CHECK(oe_zero_matrix(&dAh_dlna));
    OE_CHECK(oe_zero_matrix(&dAv_dlna));

    if (nhailgates) {
      OE_CHECK(oe_reset_matrix(&dAh_dhf, nraingates, nraingates));
      OE_CHECK(oe_reset_matrix(&dAv_dhf, nraingates, nraingates));
      OE_CHECK(oe_zero_matrix(&dAh_dhf));
      OE_CHECK(oe_zero_matrix(&dAv_dhf));
      if (scan->is_Zdr) {
	OE_CHECK(oe_reset_matrix(&dZdr_dhf, nraingates, nraingates));
	OE_CHECK(oe_zero_matrix(&dZdr_dhf));
      }
      if (scan->is_phi_dp) {
	OE_CHECK(oe_reset_matrix(&dphi_dp_dhf, nraingates, nraingates));
	OE_CHECK(oe_zero_matrix(&dphi_dp_dhf));
      }
    }

    /* Subset matrices where appropriate */
    OE_CHECK(oe_subset_matrix(&W, &W_full, &gate_list, 0, nb-1));
    if (scan->is_Zdr) {
      OE_CHECK(oe_link_matrix(&dZdr_dlna, nraingates, nraingates,
			      Jacobian_highres.data));
    }
    if (scan->is_phi_dp) {
      OE_CHECK(oe_link_matrix(&dphi_dp_dlna, nraingates, nraingates,
		      Jacobian_highres.data + nraingates*scan->is_Zdr));
    }
    OE_CHECK(oe_link_matrix(&Jacobian_rainonly, ny, nb, Jacobian.data));

    if (nb > 2 && settings->explicit_range_smoothing) {
      /* Not yet checked if the following formulation for
	 decorrelation factor is appropriate when using a smoothness
	 constraint... */
      OEreal radial_decorr_factor = settings->radial_decorr_dist*nb
	/(drange*ngates); 
      OE_CHECK(oe_reset_smoothness_constraint(nx));
      OE_CHECK(oe_register_smoothness_constraint(radial_decorr_factor,
						 0, nb-1));
      //      OE_CHECK(oe_register_flatness_constraint(settings->range_decorr_factor,
      //						 0, nb-1));
    }
    else if (nhailgates) {
      OE_CHECK(oe_reset_smoothness_constraint(nx));
    }
    else {
      OE_CHECK(oe_reset_smoothness_constraint(0));
    }

    if (nhailgates) {
      int ih, ih1 = 0, ih2 = 0, ihailgate1 = 0, ihailgate2 = -1;
      for (ih = 0; ih < nhailgates; ih++) {
	int ihailgate = retrieval->hail_location.data[iray][ih+1];
	if (ihailgate2 < 0) {
	  /* First hail gate */
	  ihailgate1 = ihailgate2 = ihailgate;
	  ih1 = ih2 = ih;
	}
	else if (ihailgate > ihailgate2+1 || ih == nhailgates-1) {
	  /* A gap or the final hail gate: smooth the previous
	     contiguous region */
	  if (ih == nhailgates-1) {
	    ih2 = ih;
	  }
	  if (ihailgate2 > ihailgate1+1) {
	    OE_CHECK(oe_register_pinned_smoothness_constraint(200.0,
							      nb+na+ih1,
							      nb+na+ih2));
	  }
	  ihailgate1 = ihailgate2 = ihailgate;
	  ih1 = ih2 = ih;
	}
	else {
	  ihailgate2 = ihailgate;
	  ih2 = ih;
	}
      }
    }

    if (settings->explicit_range_smoothing) {
      OE_CHECK(oe_register_diagonal_a_priori_invs_i(A_PRIORI, &xa, &inv_sa));
    }
    else {
      OE_CHECK(oe_register_a_priori_invs_i(A_PRIORI, &xa, &inv_Sa));
    }

    if (npass > 1 && invHessian.n > 0) {
      /* Add Kalman terms */
      OEmatrix invSlna = OE_NEW_MATRIX;
      OE_CHECK(oe_reset_matrix(&invSkalman, nx, nx));
      OE_CHECK(oe_zero_matrix(&invSkalman));
      OE_CHECK(oe_link_matrix(&invSlna, nb+na, nb+na, invSkalman.data));
      OE_CHECK(oe_reset_vector(&xkalman, nx));
      OE_CHECK(oe_zero_vector(&xkalman));
      /* Just use the basis-function part if there was hail in
	 previous ray */
      invHessian.m = invHessian.n = nb+na;
      OE_CHECK(add_spatial_decorrelation(&invHessian,
					 settings->azim_decorr_dist,
					 settings->var_ln_a_apriori,
					 &scan->range,
					 scan->azim.data[iray], last_azim));
      OE_CHECK(oe_copy_vector_part(&ln_a_basis, 0, nb, xkalman.data));
      if (na) {
	xkalman.data[nb] = atten_scaling;
	invHessian.data[nb][nb] += settings->atten_scaling_decorr;
      }
      OE_CHECK(INVERT(&invHessian, &invSlna));
      //      OE_CHECK(oe_print_row_vector(stderr, &xkalman, "% xkalman"));
      //      OE_CHECK(oe_register_a_priori_i(KALMAN_FORWARD, &x, &invHessian));
      /*
      OE_CHECK(oe_reset_vector(&sp, invHessian.n));
      for (ix = 0; ix < nx; ix++) {
	sp.data[ix] = invHessian.data[ix][ix];
      }
      OE_CHECK(oe_register_diagonal_a_priori_i(KALMAN_FORWARD, &x, &sp));
      */
      //      OE_CHECK(oe_print_a_priori_status(stderr));
      
      OE_WARN(oe_register_a_priori_invs_i(KALMAN_FORWARD,
					  &xkalman, &invSkalman));
      if (oe_error_code) {
	fprintf(stderr, "Saving erroneous inverse error covariance\n");
	OE_CHECK(oe_save_matrix("invSx_error.dat", &invSkalman, "% invSx"));
	exit(oe_error_code);
      }
      
      if (ipass > 0 && iray > 0) {
	OEmatrix Slna = OE_NEW_MATRIX;
	OE_CHECK(oe_link_matrix(&Slna, nb+na, nb+na,
				Slna_save.data + (iray-1)*(nb+na)));
	OE_CHECK(add_spatial_decorrelation(&Slna,
					   settings->azim_decorr_dist,
					   settings->var_ln_a_apriori,
					   &scan->range,
					   scan->azim.data[iray],
					   scan->azim.data[iray-1]));
	if (na) {
	  Slna.data[nb][nb] += settings->atten_scaling_decorr;
	}
	//	OE_CHECK(oe_print_matrix(stderr, &Slna_save, "Slna_save"));
	OE_CHECK(INVERT(&Slna, &invSlna));
	memcpy(xkalman.data, lna_save.data[iray], (nb+na)*sizeof(OEreal));
	OE_CHECK(oe_register_a_priori_invs_i(KALMAN_BACKWARD,
					     &xkalman, &invSkalman));
      }
      else {
	OE_CHECK(oe_register_a_priori_i(KALMAN_BACKWARD, NULL, NULL));
      }
    }
    else {
      OE_CHECK(oe_register_a_priori_i(KALMAN_FORWARD, NULL, NULL));
      OE_CHECK(oe_register_a_priori_i(KALMAN_BACKWARD, NULL, NULL));
    }

    OE_CHECK(oe_reset_vector(&x, nx));
    OE_CHECK(oe_reset_matrix(&Hessian, nx, nx));
    OE_CHECK(oe_reset_matrix(&invHessian, nx, nx));
    /* First guess */
    /* First the ln(a) values */
    for (ix = 0; ix < nb; ix++) {
      x.data[ix] = ln_a_default;
    }
    if (na) {
      x.data[ix] = settings->atten_scaling;
      ix++;
    }
    /* Then the initial hail fraction values */
    for ( ; ix < nx; ix++) {
      x.data[ix] = 0.0;
    }

    sec_reset += oe_time_split();

    //    fprintf(stderr, "Iterating");
    for (iter = 0; iter < iter_max; iter++) {
      //      oe_print_row_vector(stderr, &x, "% x");
      //      fprintf(stderr, ".");

      oe_time_split();
      if (na) {
	atten_scaling = x.data[nb];
      }
      OE_CHECK(oe_link_vector(&ln_a_basis, nb, x.data));
      OE_CHECK(oe_band_diagonal_x_vector(&W, &ln_a_basis, &ln_a));
      sec_spline1 += oe_time_split();      

      if (nhailgates) {
	/* Forward model with hail */
	/* First copy hail data from x to hail_fraction */
	for (ihailgate = 0; ihailgate < nhailgates; ihailgate++) {
	  hail_fraction.data[hail_gate_list.data[ihailgate]]
	    = x.data[nb+na+ihailgate];
	}

	/* Now run forward model */
	OE_CHECK(forward_model(lut, nraingates, drange, range.data,
			       temperature.data, Zh_lin.data, ln_a.data,
			       settings->b, hail_fraction.data,
			       atten_scaling, settings->max_atten,
			       /* Output */
			       Zdr_fwd.data, phi_dp_fwd.data, Ah.data, Av.data,
			       dZdr_dlna.data, dphi_dp_dlna.data,
			       dAh_dlna.data, dAv_dlna.data,
			       dZdr_dhf.data, dphi_dp_dhf.data,
			       dAh_dhf.data, dAv_dhf.data,
			       dZdr_das.data, dphi_dp_das.data,
			       dAh_das.data, dAv_das.data));
      }
      else {
	/* Forward model without hail */
	OE_CHECK(forward_model(lut, nraingates, drange, range.data,
			       temperature.data, Zh_lin.data, ln_a.data,
			       settings->b, NULL,
			       atten_scaling, settings->max_atten,
			       /* Output */
			       Zdr_fwd.data, phi_dp_fwd.data, Ah.data, Av.data,
			       dZdr_dlna.data, dphi_dp_dlna.data,
			       dAh_dlna.data, dAv_dlna.data,
			       NULL, NULL, NULL, NULL,
			       dZdr_das.data, dphi_dp_das.data,
			       dAh_das.data, dAv_das.data));
      }

      sec_fm += oe_time_split();

      /* Note that Jacobian_rainonly is a subset of the Jacobian
	 matrix */
      OE_CHECK(oe_matrix_x_band_diagonal(&Jacobian_highres, &W,
					 &Jacobian_rainonly));
      if (na) {
	/* Copy relevant derivatives with respect to attenuation
	   scaling into Jacobian */
	if (scan->is_Zdr) {
	  for (iraingate = 0; iraingate < nraingates; iraingate++) {
	    Jacobian.data[iraingate][nb] = dZdr_das.data[iraingate];
	  }
	}
	if (scan->is_phi_dp) {
	  for (iraingate = 0; iraingate < nraingates; iraingate++) {
	    Jacobian.data[iraingate+nraingates*scan->is_Zdr][nb]
	      = dphi_dp_das.data[iraingate];
	  }
	}
      }

      if (nhailgates) {
	/* Copy relevant parts of derivatives with respect to hail
	   fraction into Jacobian */
	if (scan->is_Zdr) {
	  for (iraingate = 0; iraingate < nraingates; iraingate++) {
	    for (ihailgate = 0; ihailgate < nhailgates; ihailgate++) {
	      Jacobian.data[iraingate][nb+na+ihailgate]
		= dZdr_dhf.data[iraingate][hail_gate_list.data[ihailgate]];
	    }
	  }
	}
	if (scan->is_phi_dp) {
	  for (iraingate = 0; iraingate < nraingates; iraingate++) {
	    for (ihailgate = 0; ihailgate < nhailgates; ihailgate++) {
	      Jacobian.data[iraingate+nraingates*scan->is_Zdr][nb+na+ihailgate]
		= dphi_dp_dhf.data[iraingate][hail_gate_list.data[ihailgate]];
	    }
	  }
	}
      }

      //      if (nhailgates > 10 && iter == 1) {

      if (0) {
      if (iray == 94 && nhailgates > 10) {
	OE_CHECK(oe_save_matrix("Jacobian.dat", &Jacobian, "% Jacobian"));
	OE_CHECK(oe_save_matrix("Jacobian_highres.dat", &Jacobian_highres, "% Jacobian"));
	/*
	OE_CHECK(oe_save_matrix("dZdr_das.dat", &dZdr_das, "% dZdr_das"));
	OE_CHECK(oe_save_matrix("dphi_dp_das.dat", &dphi_dp_das, "% dphi_dp_das"));
	OE_CHECK(oe_save_matrix("dAh_das.dat", &dAh_das, "% dAh_das"));
	OE_CHECK(oe_save_matrix("dZdr_dlna.dat", &dZdr_dlna, "% dZdr_dlna"));
	OE_CHECK(oe_save_matrix("dphi_dp_dlna.dat", &dphi_dp_dlna, "% dphi_dp_dlna"));
	OE_CHECK(oe_save_matrix("dAh_dlna.dat", &dAh_dlna, "% dAh_dlna"));
	*/
	OE_THROW(OE_FAILURE,"Quitting");
      }
      }
      sec_spline2 += oe_time_split();

      OE_CHECK(oe_prepare(&x, &y, &y_fwd, &Jacobian, &inv_sy, &Hessian, &rhs));

      sec_prep += oe_time_split();

      if (iter == iter_max-1) {
	break;
      }
      OE_WARN(SOLVE(&Hessian, &rhs));
      if (oe_error_code) {
	sec_solve += oe_time_split();
	fprintf(stderr, "Error at ray %d iteration %d\n"
		"Saving erroneous Hessian and continuing\n",
		iray, iter);
	OE_CHECK(oe_save_matrix("Hessian_error.dat", &Hessian, "% Hessian"));
	continue;
      }
      sec_solve += oe_time_split();
      //      fprintf(stderr, "\n%g %g %g", x.data[0], x.data[1], x.data[2]);
      for (ix = 0; ix < nx; ix++) {
	x.data[ix] += rhs.data[ix];
      }
      /* Check that "a" is within bounds */
      for (ix = 0; ix < nb; ix++) {
	if (x.data[ix] < xa.data[ix] - settings->max_ln_a_excursion) {
	  x.data[ix] = xa.data[ix] - settings->max_ln_a_excursion;
	}
	else if (x.data[ix] > xa.data[ix] + settings->max_ln_a_excursion) {
	  x.data[ix] = xa.data[ix] + settings->max_ln_a_excursion;
	}
      }

      if (na) {
	if (x.data[nb] < 0.2) {
	  x.data[nb] = 0.2;
	}
	else if (x.data[nb] > 5.0) {
	  x.data[nb] = 5.0;
	}
      }
      if (nhailgates) {
	/* Check that hail fraction is within bounds */
	for (ix = nb+na; ix < nx; ix++) {
	  if (x.data[ix] < 0.00001) {
	    x.data[ix] = 0.00001;
	  }
	  else if (x.data[ix] > 0.99) {
	    /* Use 0.99 in case of division by zero somewhere */
	    x.data[ix] = 0.99;
	  }
	}
      }
    }
    /* Finished iterations */
    //    if (na) {
    //      fprintf(stderr, "%g ", x.data[nb]);
    //    }

    //    OE_CHECK(oe_chi_squared(&y, &y_fwd, &inv_sy, &chi_squared));
    if (ipass+1 == npass) {
      /* If this is the backward (second) pass, calculate chi-squared
	 for Zdr and phi_dp */
      if (scan->is_Zdr) {
	for (iraingate = 0; iraingate < nraingates; iraingate++) {
	  chi_squared_Zdr +=  inv_sy.data[iraingate]
	    * (y.data[iraingate]-y_fwd.data[iraingate])
	    * (y.data[iraingate]-y_fwd.data[iraingate]);
	}
      }
      if (scan->is_phi_dp) {
	for (iraingate = nraingates*scan->is_Zdr;
	     iraingate < nraingates*(1+scan->is_Zdr); iraingate++) {
	  chi_squared_phi_dp += inv_sy.data[iraingate]
	    * (y.data[iraingate]-y_fwd.data[iraingate])
	    * (y.data[iraingate]-y_fwd.data[iraingate]);
	}
      }
      n_total += nraingates;
    }
    //    OE_CHECK(oe_print_row_vector(stderr, &inv_sy, "% inv_sy"));

    //    fprintf(stderr, "%g(%g) ", chi_squared, chi_squared/y.n);

    //    OE_CHECK(oe_save_matrix("Hessian.dat", &Hessian, "% Hessian"));
    //    OE_CHECK(oe_invert_symmetric(&Hessian));
    OE_CHECK(INVERT(&Hessian, &invHessian));
    /* Save the state if doing Kalman filtering */
    if (npass > 1) {
      OE_CHECK(oe_copy_vector_part(&x, 0, nb+na, lna_save.data[iray]));
      //      fprintf(stderr, "Copying invHessian to Sln_save: %d %d %d\n", iray, nb, na);
      OE_CHECK(oe_copy_matrix_part(&invHessian, 0, 0, nb+na, nb+na,
				   &Slna_save, iray*(nb+na), 0));
      //      if (iray == 200) {
      //	oe_print_matrix(stderr, &Slna_save, "Slna_save at 200");
      //      }
    }

    if (ipass == npass-1 || iray == nrays-1) {
      /* Put the solution in the retrieval structure */
      igate = 0;
      iraingate = 0;
      /* Do the hail data */
      if (nhailgates) {
	for (ihailgate = 0; ihailgate < nhailgates; ihailgate++) {
	  int igate = retrieval->hail_location.data[iray][ihailgate+1];
	  /* Define hail_fraction as the fraction of the Z due to hail,
	     rather than the reflectivity of the hail. If this is
	     changed then the use of it below to work out the rain
	     rate will also need to be changed. */
	  retrieval->hail_fraction.data[iray][igate]
	    = x.data[nb+na+ihailgate];// * pow(10.0, 0.1*scan->Zh.data[iray][igate]);
	}
      }
      while (igate < ngates) {
	if (iraingate < nraingates && gate_list.data[iraingate] == igate) {
	  /* A rainy gate */
	  if (scan->is_Zdr) {
	    retrieval->Zdr_fwd.data[iray][igate] = Zdr_fwd.data[iraingate];
	  }
	  if (scan->is_phi_dp) {
	    last_phi_dp = retrieval->phi_dp_fwd.data[iray][igate]
	      = phi_dp_fwd.data[iraingate];
	  }
	  last_Ah = retrieval->Ah.data[iray][igate] = Ah.data[iraingate];
	  last_Av = retrieval->Av.data[iray][igate] = Av.data[iraingate];
	  if (nhailgates) {
	    retrieval->R.data[iray][igate] 
	      = pow( (1.0-retrieval->hail_fraction.data[iray][igate])
		     * Zh_lin.data[iraingate]
		     * pow(10.0, 0.1*Ah.data[iraingate])
		     * exp(-ln_a.data[iraingate]),
		    1/settings->b);
	  }
	  else {
	    retrieval->R.data[iray][igate] 
	      = pow(Zh_lin.data[iraingate] * pow(10.0, 0.1*Ah.data[iraingate])
		    * exp(-ln_a.data[iraingate]), 1.0/settings->b);
	    //	  retrieval->lnR_error.data[iray][igate]
	    //	    = invHessian.data[iraingate][iraingate]/settings->b;
	  }
	  iraingate++;
	}
	else {
	  /* A non-rainy gate */
	  retrieval->Zdr_fwd.data[iray][igate] = last_Av-last_Ah;
	  retrieval->R.data[iray][igate] = 0.0;
	  //	  retrieval->lnR_error.data[iray][igate] = 0.0;
	  retrieval->phi_dp_fwd.data[iray][igate] = last_phi_dp;
	  retrieval->Ah.data[iray][igate] = last_Ah;
	  retrieval->Av.data[iray][igate] = last_Av;
	}
	igate++;
      }
      /* Now do "a" */
      OE_CHECK(oe_link_vector(&ln_a_full, ngates, retrieval->a.data[iray]));
      OE_CHECK(oe_band_diagonal_x_vector(&W_full, &ln_a_basis, &ln_a_full));

      /* Exponentiate */
      for (igate = 0; igate < ngates; igate++) {
	ln_a_full.data[igate] = exp(ln_a_full.data[igate]);
      }

      if (iray == iray_save) {
	OE_CHECK(oe_save_matrix("Hessian_ray.dat", &Hessian, "% Hessian"));
	OE_CHECK(oe_save_matrix("Jacobian_ray.dat", &Jacobian, "% Jacobian"));
	OE_CHECK(oe_save_vector("x_ray.dat", &x, "% x"));
	OE_CHECK(oe_save_vector("dAh_das_ray.dat", &dAh_das, "% dAh_das"));
      }

      /* Do the error in the natural logarithm of rainrate */
      OE_CHECK(oe_subtract_a_priori(&Hessian,
				    KALMAN_FORWARD, KALMAN_BACKWARD));
      OE_CHECK(oe_reset_matrix(&Sa_true, Hessian.m, Hessian.n));
      OE_CHECK(INVERT(&Hessian, &Sa_true));      
      if (iray == iray_save) {
	OE_CHECK(oe_save_matrix("Sa_ray.dat", &Sa_true, "% Sa"));
      }

      /* We are only interested in the rainy part of the inverse
	 Hessian */
      OE_CHECK(oe_link_vector(&lnR_error, retrieval->lnR_error.n,
			      retrieval->lnR_error.data[iray]));
      OE_CHECK(oe_zero_vector(&lnR_error));
      Sa_true.m = Sa_true.n = nb;
      OE_CHECK(oe_sparse_W_invA_WT_diagonal(&W_full, &Sa_true,
					    &lnR_error));
      Sa_true.m = Sa_true.n = nx;
      /* At this point, lnR_error contains the error variance of
	 ln(a) */
      for (igate = 0; igate < ngates; igate++) {
	retrieval->lna_error.data[iray][igate] = sqrt(lnR_error.data[igate]);
      }

      if (scan->is_Zh_error) {
	/* If Zh_error is present then use it */
	for (igate = 0; igate < ngates; igate++) {
	  /* Note that 0.053019 = [ln(10)/10]^2 */
	  lnR_error.data[igate] += 0.053019
	    * scan->Zh_error.data[iray][igate]
	    * scan->Zh_error.data[iray][igate];
	}
      }
      if (na) {
	OEreal factor = 0.053019 * Sa_true.data[nb][nb]
	  / (atten_scaling*atten_scaling);
	/* If attenuation scaling is present then use it */
	for (iraingate = 0; iraingate < nraingates; iraingate++) {
	  int igate = gate_list.data[iraingate];
	  lnR_error.data[igate] += factor
	    * dAh_das.data[iraingate]
	    * dAh_das.data[iraingate];
	}
      }
      /* Do the hail contribution to the rain error*/
      if (nhailgates) {
	for (ihailgate = 0; ihailgate < nhailgates; ihailgate++) {
	  int igate = retrieval->hail_location.data[iray][ihailgate+1];
	  lnR_error.data[igate]
	    += Sa_true.data[nb+na+ihailgate][nb+na+ihailgate]
	    / ((1.0-retrieval->hail_fraction.data[iray][igate])
	       * (1.0-retrieval->hail_fraction.data[iray][igate]));
	}
      }

      /* Square-root and scale to obtain the actual error in ln(R) */
      for (igate = 0; igate < ngates; igate++) {
	lnR_error.data[igate] = sqrt(lnR_error.data[igate])/settings->b;
      }

      /* Do the attenuation scaling */
      if (na) {
	retrieval->atten_scaling.data[iray] = atten_scaling;
	retrieval->atten_scaling_error.data[iray] = sqrt(Sa_true.data[nb][nb]);
      }
      /* Do the hail error */
      if (nhailgates) {
	for (ihailgate = 0; ihailgate < nhailgates; ihailgate++) {
	  int igate = retrieval->hail_location.data[iray][ihailgate+1];
	  retrieval->hail_fraction_error.data[iray][igate]
	    = sqrt(Sa_true.data[nb+na+ihailgate][nb+na+ihailgate]);
	}
      }

      /*
	OE_CHECK(oe_save_matrix("Jacobian.dat", &Jacobian, "% Jacobian"));
	OE_CHECK(oe_save_matrix("Jacobian_highres.dat", &Jacobian_highres, "% Jacobian_highres"));
	exit(1);
      */
    }
    sec_save += oe_time_split();

    last_azim = scan->azim.data[iray];
    
    if (ipass == 0) {
      /* First pass: move to next ray */
      if (++iray >= nrays) {
	ipass++;
	iray -= 2;
	if (npass > 1) {
	  fprintf(stderr, "\nBackward pass ");
	}
      }      
    }
    else {
      /* Second pass: move to previous ray */
      if (--iray < 0) {
	ipass++;
      }
    }
  }
  fprintf(stderr, "\n");

  if (scan->is_Zdr) {
    fprintf(stderr, "Zdr chi-squared = %g (n = %d, rms = %g)\n",
	    chi_squared_Zdr, n_total, sqrt(chi_squared_Zdr/n_total));
  }
  if (scan->is_Zdr) {
    fprintf(stderr, "phi_dp chi-squared = %g (n = %d, rms = %g)\n",
	    chi_squared_phi_dp, n_total, sqrt(chi_squared_phi_dp/n_total));
  }


  fprintf(stderr, "Time spent (seconds)\n");
  fprintf(stderr, "   resetting matrices:         %g\n", sec_reset);
  fprintf(stderr, "   initial spline conversion:  %g\n", sec_spline1);
  fprintf(stderr, "   forward model:              %g\n", sec_fm);
  fprintf(stderr, "   spline Jacobian conversion: %g\n", sec_spline2);
  fprintf(stderr, "   matrix preparation:         %g\n", sec_prep);
  fprintf(stderr, "   matrix solution:            %g\n", sec_solve);
  fprintf(stderr, "   saving result:              %g\n", sec_save);

  oe_free_vector(&range);
  oe_free_vector(&Zh_lin);
  oe_free_vector(&temperature);

  oe_free_vector(&y);
  oe_free_vector(&inv_sy);

  oe_free_vector(&y_fwd);
  oe_free_vector(&Zdr_fwd);
  oe_free_vector(&phi_dp_fwd);

  oe_free_vector(&x);
  oe_free_vector(&ln_a);
  oe_free_vector(&ln_a_full);

  oe_free_vector(&xa);
  if (settings->explicit_range_smoothing) {
    oe_free_vector(&inv_sa);
  }
  else {
    oe_free_matrix(&inv_Sa);
  }

  oe_free_matrix(&W_full);
  oe_free_matrix(&W);

  oe_free_vector(&Ah);
  oe_free_vector(&Av);

  oe_free_matrix(&Jacobian);

  oe_free_matrix(&Jacobian_highres);
  oe_free_matrix(&dZdr_dlna);
  oe_free_matrix(&dphi_dp_dlna);

  oe_free_matrix(&Hessian);
  oe_free_matrix(&invHessian);
  oe_free_matrix(&Sa_true);
  oe_free_vector(&rhs);
  oe_free_vector(&sp);

  oe_free_matrix(&dAh_dlna);
  oe_free_matrix(&dAv_dlna);

  oe_free_matrix(&dAh_dhf);
  oe_free_matrix(&dAv_dhf);
  oe_free_matrix(&dZdr_dhf);
  oe_free_matrix(&dphi_dp_dhf);

  if (na) {
    oe_free_vector(&dAh_das);
    oe_free_vector(&dAv_das);
    oe_free_vector(&dZdr_das);
    oe_free_vector(&dphi_dp_das);
  }

  oe_free_matrix(&lna_save);
  oe_free_matrix(&Slna_save);

  oe_free_int_vector(&gate_list);

  return OE_SUCCESS;
}
